"""Base/interface of each module."""
from abc import ABC
from PySide2 import QtGui
from PySide2.QtCore import Qt


def qfont_with_pixel_size(font_family, pixel_size, weight=None):
    """Provide a QFont with given family and pixel size.

    Created because QFont does not have a constructor with pixel size as a
    parameter.

    Parameters
    ----------
    font_family : str
        Name of font family. Sent to https://doc.qt.io/qt-5/qfont.html#QFont-1
    pixel_size : int
        Pixel size. Sent to https://doc.qt.io/qt-5/qfont.html#setPixelSize
    weight : QFont.Weight
        Weight of font. Sent to https://doc.qt.io/qt-5/qfont.html#QFont-1
    """
    if weight is not None:
        font = QtGui.QFont(font_family, weight=weight)
    else:
        font = QtGui.QFont(font_family)
    font.setPixelSize(pixel_size)
    return font


# Font for module headline
HEADLINE_FONT = qfont_with_pixel_size("Arial", 24, 100)

TABLE_CELL_DEFAULT_FONT = qfont_with_pixel_size("Arial", 16)


class ModuleBase(ABC):
    """Interface for SOI-modules."""

    type = None

    def __init__(self):
        """Class-variable 'type' should be set by derived class."""
        if self.type is None:
            raise NotImplementedError

    def get_size(self):
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    def get_data(self):
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    @staticmethod
    def get_user_friendly_name():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    @staticmethod
    def get_icon():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError


def resize_table(table, rows=True, columns=True, has_headline=False):
    """Resize a given QTableWidget.

    Parameters
    ----------
    table : QTableWidget
        QTablewidget-instance to resize.
    rows : bool
        Resizes rows to contents if True, by default True.
    columns : bool
        Resizes columns to contents if True, by default True.
    has_headline : bool
        True if the table has a headline, by default False.
        Last column is widened if headline is wider than table.
    """
    if columns:
        table.resizeColumnsToContents()
    if rows:
        table.resizeRowsToContents()

    # If table has a headline, make sure table is wide enough to fit it.
    if has_headline:
        last_column_index = table.columnCount() - 1
        table.resizeColumnToContents(last_column_index)
        width, height = get_table_size(table)

        # Get width of headline
        headline = table.item(0, 0).text()
        headline_width = (
            QtGui.QFontMetricsF(HEADLINE_FONT).horizontalAdvance(headline) + 10
        )
        # If headline is wider than table
        if width < headline_width:
            difference = headline_width - width
            width += difference
            old_width = table.columnWidth(last_column_index)
            table.setColumnWidth(last_column_index, old_width + difference)
    else:
        width, height = get_table_size(table)

    table.setFixedWidth(width)
    table.setFixedHeight(height)


def get_table_size(widget):
    """Calculate and return total width and height of a QTableWidget.

    Parameters
    ----------
    widget : QTableWidget
        QTableWidget-instance to calculate and return size of.

    Returns
    -------
    Tuple
        Total (width, height)
    """
    width = 0
    height = 0

    for i in range(widget.columnCount()):
        width += widget.columnWidth(i) + 0.5

    for i in range(widget.rowCount()):
        height += widget.rowHeight(i) + 0.5

    return width, height


def set_module_pos(widget, pos):
    """Set position of module (widget).

    Parameters
    ----------
    widget : QWidget
        Widget to move.
    pos : QPoint
        Position (x, y).
    """
    widget.move(pos)


def event_is_ctrl_plus(event):
    """Check if the event is 'CTRL +'.

    Parameters
    ----------
    event : QKeyEvent
        Event to check.
    """
    return (
        event.modifiers() == Qt.ControlModifier and event.key() == Qt.Key_Plus
    )


def event_is_ctrl_minus(event):
    """Check if the event is 'CTRL -'.

    Parameters
    ----------
    event : QKeyEvent
        Event to check.
    """
    # Underline is in practice minus in this situation. We don't know why
    return (
        event.modifiers() == Qt.ControlModifier
        and event.key() == Qt.Key_Underscore
    )


def event_is_shift_plus(event):
    """Check if the event is 'SHIFT +'.

    Parameters
    ----------
    event : QKeyEvent
        Event to check.
    """
    return event.key() == Qt.Key_Question


def event_is_shift_minus(event):
    """Check if the event is 'SHIFT -'.

    Parameters
    ----------
    event : QKeyEvent
        Event to check.
    """
    # Underline is in practice minus in this situation. We don't know why
    return (
        event.modifiers() == Qt.ShiftModifier
        and event.key() == Qt.Key_Underscore
    )
